{% if fluff -%}/**********************************************************************************************
 * Filename:       app_snippets.c
 *
 * Description:    This file contains snippets needed to utilize the generated services.
 *
 *                 Generated by:
 *                 BDS version: {{pluginInfo.AppVersion}}
 *                 Plugin:      {{pluginInfo.Name}} {{ pluginInfo.Version}}
 *                 Time:        {{pluginInfo.timestamp}}
 *
 * Copyright (c) 2015, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************************************/
{% endif -%}
//...
#include <ti/sysbios/knl/Semaphore.h>
#include <ti/sysbios/knl/Queue.h>

//...
{% for service in services -%}
#include "{{service.name}}.h"
{% endfor -%}

//...

{% if anyWriteable -%}
// Semaphore globally used to post events to the application thread
static ICall_Semaphore sem; // Already present in sample applications.

// Queue object used for service messages.
static Queue_Struct serviceMsgQ;
static Queue_Handle hServiceMsgQ;

// Struct for messages from a service
typedef struct
{
  Queue_Elem _elem;
  uint16_t svcUUID;
  uint16_t dataLen;
  uint8_t  paramID;
  uint8_t  data[]; // Flexible array member, extended to malloc - sizeof(.)
} server_char_write_t;
{% endif -%}

//...

// Declaration of service callback handlers
{% for service in services -%}
{% if service.writeable -%}
static void user_{{service.nameCap}}_ValueChangeCB(uint8_t paramID); // Callback from the service.
static void user_{{service.nameCap}}_ValueChangeDispatchHandler(server_char_write_t *pWrite); // Local handler called from the Task context of this task.
{% else -%}
// {{service.name}} has no writeable characteristics
{% endif -%}
{% endfor -%}

// Service callback structure
{% for service in services -%}{% if service.writeable -%}
// {{ service.nameCap }} callback handler. The type {{service.name}}CBs_t is defined in {{service.name}}.h
static {{service.nameCap}}CBs_t user_{{service.name}}CBs =
{
  user_{{service.nameCap}}_ValueChangeCB // Characteristic value change callback handler
};
{% else -%}
// {{service.name}} has no writeable characteristics.
{% endif -%}
{% endfor -%}


// taskFxn_init(..)
// {
{% if anyWriteable -%}
  // Initialize queue for service messages.
  // Note: Used to transfer control to application thread
  Queue_construct(&serviceMsgQ, NULL);
  hServiceMsgQ = Queue_handle(&serviceMsgQ);
{% endif -%}

{% for service in services -%}
  {{ service.nameCap }}_AddService();
{% if service.writeable -%}
  {{ service.nameCap }}_RegisterAppCBs(&user_{{service.name}}CBs);
{% endif -%}
{% endfor -%}

  // Placeholder variable for characteristic intialization
  uint8_t someVal[20] = {0};

{% for service in services -%}
  // Initalization of characteristics in {{service.name}} that are readable.
{% for char in service.charsReadable -%}
  {{ service.nameCap }}_SetParameter({{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID, {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_LEN, &someVal);
{% endfor -%}

{% endfor -%}
//}


// taskFxn(..)
// {
//   for (;;) {
//     ... ICall_wait(...)   // Wait for semaphore post
// 
{% if anyWriteable -%}
      // Process messages sent from another task or another context.
      while (!Queue_empty(hServiceMsgQ))
      {
        server_char_write_t *pWrite = Queue_dequeue(hServiceMsgQ);

        switch(pWrite->svcUUID)
        {
{% for service in services -%}
{% if service.writeable -%}
          case {{ service.name | upcase }}_SERV_UUID:
            user_{{service.nameCap}}_ValueChangeDispatchHandler(pWrite);
            break;
{% endif -%}
{% endfor -%}
        }

        // Free the message received from the service callback.
        ICall_free(pWrite);
      }
{% else -%}
//   No writeable characteristics, so queue handling is not included.
{% endif -%}

//   }
// }

{% for service in services -%}
{% if service.writeable -%}

/**
 * Handle received write that was sent as message to the application task.
 */
void user_{{service.nameCap}}_ValueChangeDispatchHandler(server_char_write_t *pWrite)
{
  switch (pWrite->paramID)
  {
{% for char in service.charsNeedWrite -%}
    case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
      // Do something useful with pWrite->data here
      // -------------------------
      break;
{% endfor -%}
  }

}
{% endif -%}
{% endfor -%}

{% for service in services -%}
{% if service.writeable -%}

/**
 * Callback handler from the service to application when something is written
 * to a writeable characteristic, called from Stack task context.
 */
static void user_{{service.nameCap}}_ValueChangeCB(uint8_t paramID)
{
  // See {{service.name}}.h to compare paramID with characteristic value attribute.
  // Called in Stack Task context, so can't do processing here.

  // Send message to application message queue about received data.
  uint16_t readLen = 0; // How much to read via service API
  
  switch (paramID)
  {
{% for char in service.charsNeedWrite -%}
    case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
      readLen = {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_LEN;
      break;
{% endfor -%}
  }

  // Allocate memory for the message.
  // Note: The pWrite message doesn't have to contain the data itself, as that's stored in
  //       a variable in the service. However, to prevent data loss if a new value is received
  //       before GetParameter is called, we call GetParameter now.
  // Note: It could be just as well to send dataLen and a pointer to the received data directly to this callback, avoiding GetParameter alltogether.
  server_char_write_t *pWrite = ICall_malloc(sizeof(server_char_write_t) + readLen);

  if (pWrite != NULL)
  {
    pWrite->svcUUID = {{ service.name | upcase }}_SERV_UUID;
    pWrite->paramID = paramID;
    // Get the data from the service API. readLen is now equal to allocated space.
    {{ service.nameCap }}_GetParameter( paramID, &readLen, pWrite->data );
    pWrite->dataLen = readLen; // Store length of received data after GetParameter updates readLen.
    // Enqueue the message using pointer to queue node element.
    Queue_enqueue(hServiceMsgQ, &pWrite->_elem);
    // Let application know there's a message
    Semaphore_post(sem);
  }
}
{% endif -%}
{% endfor -%}
